﻿/*---------------------------------------------------------------------------
 *
 *	Title                : EMESARY Class based inter-object communication
 *
 *	File Type            : Implementation File
 *
 *	Description          : Provides generic inter-object communication. For an object to receive a message it
 *	                     : must first register with a Transmitter, such as GlobalTransmitter, and implement the 
 *	                     : IReceiver interface. That's it.
 *	                     : To send a message use a Transmitter with an object. That's all there is to it.
 *	                     : 
 * 
 *  References           : http://www.chateau-logic.com/content/class-based-inter-object-communication
 *
 *	Author               : Richard Harrison (richard@zaretto.com)
 *
 *	Creation Date        : 24 September 2009
 *
 *	Version              : $Header: $
 *
 *  Copyright © 2009 Richard Harrison           All Rights Reserved.
 *
 *---------------------------------------------------------------------------*/

using System.Collections.Generic;

namespace Emesary
{
    public enum NotificationType // This should be more extensible - but it is sufficient for now
    {
        CustomerView_Delete = 0xee0001,
        CustomerView_CustomersSelected,
        SelectView,
        SelectMenu,  // Value contains Menu
        SelectRecipe, // Value contains Recipe
        ViewSelected, // Notification that the view has been changed - value is MMView.

        CreateMenu,
        DeleteMenu,
        CreateRecipe,
        DeleteRecipe,

        QueryCanAccept,
        ItemDropped,
        
        ApplicationCommand,
        ApplicationCommandCanExecute
    }

    public enum ReceiptStatus
    {
        OK = 0,   			// info
        Fail = 1, 			// if any item fails then send message may return fail
        Abort = 2, 			// stop processing this event and fail
        Finished = 3, 		// stop processing this event and return success
        NotProcessed = 4 	// recipient didn't recognise this event
    }

    public interface IReceiver
    {
        ReceiptStatus Receive(Notification message);
    }

    public class Notification
    {
        public Notification(NotificationType type)
        {
            this.Type = type;
            this.Value = null;
        }

        public Notification(NotificationType type, object value)
        {
            this.Type = type;
            this.Value = value;
        }

        public NotificationType Type { get ; set; }
        public object Value { get; set; }
    }

    /**
     * Description: Transmits Message derived objects. Each instance of this class provides a
     * databus to which any number of receivers can attach to. 
     *
     * Messages may be inherited and customised between individual systems.
     */
    public class Transmitter
    {
        private List<IReceiver> V = new List<IReceiver>();

        /**
         * Registers an object to receive messsages from this transmitter. 
         * This object is added to the top of the list of objects to be notified. This is deliberate as 
         * the sequence of registration and message receipt can influence the way messages are processing
         * when ReceiptStatus of Abort or Finished are encountered. So it was a deliberate decision that the
         * most recently registered recipients should process the messages/events first.
         */
        public void Register(IReceiver R)
        {
            if (V.IndexOf(R) < 0)
                V.Insert(0, R);
        }

        /*
         *  Removes an object from receving message from this transmitter
         */
        public void DeRegister(IReceiver R)
        {
            V.Remove(R);
        }

        /*
         * Notify all registered recipients. Stop when receipt status of abort or finished are received.
         * The receipt status from this method will be 
         *  - OK > message handled
         *  - Fail > message not handled. A status of Abort from a recipient will result in our status
         *           being fail as Abort means that the message was not and cannot be handled, and
         *           allows for usages such as access controls.
         * NOTE: When I first designed Emesary I always intended to have message routing and the ability
         *       for each recipient to specify an area of interest to allow performance improvements
         *       however this has not yet been implemented - but the concept is still there and
         *       could be implemented by extending the IReceiver interface to allow for this.
         */
        public ReceiptStatus NotifyAll(Notification M)
        {
            foreach (IReceiver R in V)
            {
                ReceiptStatus rstat = R.Receive(M);
                switch (rstat)
                {
                    case ReceiptStatus.Fail:
                    case ReceiptStatus.NotProcessed:
                    case ReceiptStatus.OK:
                        break;

                    case ReceiptStatus.Abort:
                        return ReceiptStatus.Fail;

                    case ReceiptStatus.Finished:
                        return ReceiptStatus.OK;
                }

            }
            return ReceiptStatus.OK;
        }
    }

    /*
     * Mainly a convenience - the global transmitter is exactly what it sounds like - an instance
     * of a transmitter that can be used globally.
     */
    public class GlobalTransmitter
    {
        private static Transmitter globalNotifier = new Transmitter();

        public static void Register(IReceiver R)
        {
            globalNotifier.Register(R);
        }
        
        public static void DeRegister(IReceiver R)
        {
            globalNotifier.DeRegister(R);
        }
        
        public static ReceiptStatus NotifyAll(Notification M)
        {
            return globalNotifier.NotifyAll(M);
        }
    }
    /*
     * Further notes: 
     *  - I first came up with this notification system in 1997 in C++ using multiple inheritance.
     *    Since then there have been various versions with differing names (PostOffice (C++) / MessageBus (Java))
     *  - There have been various features in each of the earlier versions that have never really been used such as
     *     * routing between Transmitters (by having a recipient that bridges across to another Transmitter)
     *     * area of interest filtering - the overhead of deciding which messages to handled negated the simplicity of
     *       the everyone gets everything approach
     *     * time / frequency based filtering - very limited usefulness as most messages are important and it was better
     *       to fix this at the source of message creation by issuing less messages. This was originally intended to allow
     *       avionics simulated devices to run at 30hz, 15hz and 1hz, but in fact this was better implemented by the scheduler
     *       as the coupling created by controlling the message frequency was too close.
     *     * Bridges can connect two tranmitters. A bridge is simply a recipient that relays on messages to another transmitter.
     *     * there have been process -> process bridging of messages using a bridge over a socket.
     *     * The way the message is built has changed sinced the C++ where it was a simple ID and Value, then became a templated
     *       class containing a message (which caused lots of complexity) in the Java it was a very (possibly over) 
     *       complex message containing message elements (similar to ARINC 429). 
     *       The approach taken here is the best so far - of passing an object inherited from the Notification - as this allows
     *       for quick and easy conversion and handling using RTTI / reflection.
     *       is the current favoured approach - and fits best out of
     */

    /*
     *  Contains the drag notification message. A drag notification request is
     *  a type and a source and target object.
     *  This notification may be renamed if we find other cases of notifications that
     *  require a source and target.
     */
    public class DragNotification : Notification
    {
        public object target { get { return Value; } set { Value = value; } }
        public object source { get; set; }
        public int position { get; set; }

        public DragNotification(NotificationType type, object target, object source)
            : base(type, target)
        {
            this.source = source;
            this.position = -1;
        }

        public DragNotification(NotificationType type, object target, object source, int position)
            : base(type, target)
        {
            this.source = source;
            this.position = position;
        }
    }
}
